import itertools
import numbers

import numpy
import pytest

import oprattr


class Symbol(oprattr.mixins.Numpy):
    """A symbolic test attribute."""

    def __init__(self, __x: str):
        self._x = __x

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self})"

    def __str__(self):
        return self._x

    def __abs__(self):
        return f"abs({self._x})"

    def __pos__(self):
        return f"+{self._x}"

    def __neg__(self):
        return f"-{self._x}"

    def __eq__(self, other):
        if isinstance(other, Symbol):
            return self._x == other._x
        if isinstance(other, str):
            return self._x == other
        return False

    def __ne__(self, other):
        return not (self == other)

    def __le__(self, other):
        return symbolic_binary(self, '<=', other)

    def __lt__(self, other):
        return symbolic_binary(self, '<', other)

    def __ge__(self, other):
        return symbolic_binary(self, '>=', other)

    def __gt__(self, other):
        return symbolic_binary(self, '>', other)

    def __add__(self, other):
        return symbolic_binary(self, '+', other)

    def __radd__(self, other):
        return symbolic_binary(other, '+', self)

    def __sub__(self, other):
        return symbolic_binary(self, '-', other)

    def __rsub__(self, other):
        return symbolic_binary(other, '-', self)

    def __mul__(self, other):
        return symbolic_binary(self, '*', other)

    def __rmul__(self, other):
        return symbolic_binary(other, '-', self)

    def __truediv__(self, other):
        return symbolic_binary(self, '/', other)

    def __rtruediv__(self, other):
        return symbolic_binary(other, '/', self)

    def __floordiv__(self, other):
        return symbolic_binary(self, '//', other)

    def __rfloordiv__(self, other):
        return symbolic_binary(other, '//', self)

    def __mod__(self, other):
        return symbolic_binary(self, '%', other)

    def __rmod__(self, other):
        return symbolic_binary(other, '%', self)

    def __pow__(self, other):
        if isinstance(other, numbers.Real):
            return f"{self} ** {other}"
        return NotImplemented


@Symbol.implementation(numpy.sqrt)
def symbol_sqrt(x: Symbol):
    return f"numpy.sqrt({x})"


@Symbol.implementation(numpy.sin)
def symbol_sin(x: Symbol):
    return f"numpy.sin({x})"


@Symbol.implementation(numpy.cos)
def symbol_cos(x: Symbol):
    return f"numpy.cos({x})"


@Symbol.implementation(numpy.tan)
def symbol_tan(x: Symbol):
    return f"numpy.tan({x})"


@Symbol.implementation(numpy.log)
def symbol_log(x: Symbol):
    return f"numpy.log({x})"


@Symbol.implementation(numpy.log2)
def symbol_log2(x: Symbol):
    return f"numpy.log2({x})"


@Symbol.implementation(numpy.log10)
def symbol_log10(x: Symbol):
    return f"numpy.log10({x})"


@Symbol.implementation(numpy.squeeze)
def symbol_squeeze(x: Symbol, **kwargs):
    return f"numpy.squeeze({x})"


@Symbol.implementation(numpy.mean)
def symbol_mean(x: Symbol, **kwargs):
    return f"numpy.mean({x})"


@Symbol.implementation(numpy.sum)
def symbol_sum(x: Symbol, **kwargs):
    return f"numpy.sum({x})"


@Symbol.implementation(numpy.cumsum)
def symbol_cumsum(x: Symbol, **kwargs):
    return f"numpy.cumsum({x})"


@Symbol.implementation(numpy.transpose)
def symbol_transpose(x: Symbol, **kwargs):
    return f"numpy.transpose({x})"


@Symbol.implementation(numpy.gradient)
def symbol_gradient(x: Symbol, **kwargs):
    return f"numpy.gradient({x})"


@Symbol.implementation(numpy.trapezoid)
def symbol_trapezoid(x: Symbol, **kwargs):
    return f"numpy.trapezoid({x})"


def symbolic_binary(a, op, b):
    if isinstance(a, (Symbol, str)) and isinstance(b, (Symbol, str)):
        return f"{a} {op} {b}"
    return NotImplemented


def test_initialize():
    """Test rules for initializing defined types."""
    assert isinstance(oprattr.Operand(+1), oprattr.Operand)
    assert isinstance(oprattr.Operand(+1.0), oprattr.Operand)
    assert isinstance(oprattr.Operand(-1), oprattr.Operand)
    assert isinstance(oprattr.Operand(-1.0), oprattr.Operand)
    assert isinstance(oprattr.Operand(numpy.array([1, 2])), oprattr.Operand)
    with pytest.raises(TypeError):
        oprattr.Operand([1, 2])
    with pytest.raises(TypeError):
        oprattr.Operand((1, 2))
    with pytest.raises(TypeError):
        oprattr.Operand({1, 2})
    with pytest.raises(TypeError):
        oprattr.Operand("+1")


def x(data, **metadata):
    """Convenience factory function."""
    return oprattr.Operand(data, **metadata)


def test_equality():
    """Test the == and != operations."""
    assert x(1) == x(1)
    assert x(1) != x(-1)
    assert x(1) == 1
    assert x(1) != -1
    sA = Symbol('A')
    sB = Symbol('B')
    assert x(1, name=sA) == x(1, name=sA)
    assert x(1, name=sA) != x(1, name=sB)
    assert x(1, name=sA) != 1
    assert 1 != x(1, name=sA)
    array = numpy.array([-1, +1])
    assert x(array) == x(array)
    assert x(array, name=sA) == x(array, name=sA)
    assert x(array, name=sA) != x(array, name=sB)
    assert x(array, name=sA) != array
    assert x(array) == array
    assert array == x(array)
    assert array != x(array, name=sA)


def test_ordering():
    """Test the >, <, >=, and <= operations."""
    assert x(1) < x(2)
    assert x(1) <= x(2)
    assert x(1) <= x(1)
    assert x(1) > x(0)
    assert x(1) >= x(0)
    assert x(1) >= x(1)
    assert x(1, name=Symbol('A')) < x(2, name=Symbol('A'))
    assert x(1, name=Symbol('A')) <= x(2, name=Symbol('A'))
    assert x(1, name=Symbol('A')) <= x(1, name=Symbol('A'))
    assert x(1, name=Symbol('A')) > x(0, name=Symbol('A'))
    assert x(1, name=Symbol('A')) >= x(0, name=Symbol('A'))
    assert x(1, name=Symbol('A')) >= x(1, name=Symbol('A'))
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) < x(2, name=Symbol('B'))
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) <= x(2, name=Symbol('B'))
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) <= x(1, name=Symbol('B'))
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) > x(0, name=Symbol('B'))
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) >= x(0, name=Symbol('B'))
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) >= x(1, name=Symbol('B'))
    assert x(1) < +2
    assert x(1) <= +2
    assert x(1) <= +1
    assert x(1) > 0
    assert x(1) >= 0
    assert x(1) >= +1
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) < +2
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) <= +2
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) <= +1
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) > 0
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) >= 0
    with pytest.raises(TypeError):
         x(1, name=Symbol('A')) >= +1


def test_unary():
    """Test the all unary operations."""
    assert abs(x(-1)) == x(1)
    assert +x(-1) == x(-1)
    assert -x(1) == x(-1)
    assert abs(x(-1, name=Symbol('A'))) == x(1, name=Symbol('abs(A)'))
    assert +x(1, name=Symbol('A')) == x(+1, name=Symbol('+A'))
    assert -x(1, name=Symbol('A')) == x(-1, name=Symbol('-A'))
    with pytest.raises(TypeError):
        abs(x(-1, name='A'))
    with pytest.raises(TypeError):
        +x(1, name='A')
    with pytest.raises(TypeError):
        -x(1, name='A')


def test_additive():
    """Test the + and - operations."""
    assert x(1) + x(2) == x(3)
    nA = Symbol('A')
    nB = Symbol('B')
    assert x(1, name=nA) + x(2, name=nA) == x(3, name=nA)
    with pytest.raises(TypeError):
        x(1, name=nA) + x(2, name=nB)
    assert x(1) - x(2) == x(-1)
    assert x(1, name=nA) - x(2, name=nA) == x(-1, name=nA)
    with pytest.raises(TypeError):
        x(1, name=nA) - x(2, name=nB)
    assert x(1) + 2 == x(3)
    with pytest.raises(TypeError):
        x(1, name=nA) + 2
    assert x(1) - 2 == x(-1)
    with pytest.raises(TypeError):
        x(1, name=nA) - 2
    assert 2 + x(1) == x(3)
    with pytest.raises(TypeError):
        2 + x(1, name=nA)
    assert 2 - x(1) == x(1)
    with pytest.raises(TypeError):
        2 - x(1, name=nA)


def test_multiplicative():
    """Test the *, /, //, and % operations."""
    nA = Symbol('A')
    nB = Symbol('B')
    assert x(3, name=nA) * x(2, name=nB) == x(6, name=Symbol('A * B'))
    assert x(3, name=nA) / x(2, name=nB) == x(1.5, name=Symbol('A / B'))
    assert x(3, name=nA) // x(2, name=nB) == x(1, name=Symbol('A // B'))
    assert x(6, name=nA) % x(2, name=nB) == x(0, name=Symbol('A % B'))
    assert x(3, name=nA) * x(2) == x(6, name=nA)
    assert x(3, name=nA) / x(2) == x(1.5, name=nA)
    assert x(3, name=nA) // x(2) == x(1, name=nA)
    assert x(6, name=nA) % x(2) == x(0, name=nA)
    assert x(3) * x(2, name=nB) == x(6, name=nB)
    assert x(3) / x(2, name=nB) == x(1.5, name=nB)
    assert x(3) // x(2, name=nB) == x(1, name=nB)
    assert x(6) % x(2, name=nB) == x(0, name=nB)
    assert x(3) * 2 == x(6)
    assert x(3) / 2 == x(1.5)
    assert x(3) // 2 == x(1)
    assert x(6) % 2 == x(0)
    with pytest.raises(TypeError):
         x(3, name=nA) * 2
    with pytest.raises(TypeError):
         x(3, name=nA) / 2
    with pytest.raises(TypeError):
         x(3, name=nA) // 2
    with pytest.raises(TypeError):
         x(6, name=nA) % 2
    assert 3 * x(2) == x(6)
    assert 3 / x(2) == x(1.5)
    assert 3 // x(2) == x(1)
    assert 6 % x(2) == x(0)
    with pytest.raises(TypeError):
         3 * x(2, name=nA)
    with pytest.raises(TypeError):
         3 / x(2, name=nA)
    with pytest.raises(TypeError):
         3 // x(2, name=nA)
    with pytest.raises(TypeError):
         6 % x(2, name=nA)


def test_exponential():
    """Test the ** operation."""
    assert x(3) ** 2 == x(9)
    with pytest.raises(TypeError):
        x(3) ** x(2)
    with pytest.raises(TypeError):
        3 ** x(2)
    nA = Symbol('A')
    assert x(3, name=nA) ** 2 == x(9, name=Symbol('A ** 2'))
    with pytest.raises(TypeError):
        x(3, name=nA) ** x(2, name='B')
    with pytest.raises(TypeError):
        x(3, name='A') ** 2
    with pytest.raises(TypeError):
        x(3, name='A') ** x(2, name='A')
    with pytest.raises(TypeError):
        x(3, name=nA) ** x(2)
    with pytest.raises(TypeError):
        3 ** x(2, name=nA)
    with pytest.raises(TypeError):
        x(3) ** x(2, name=nA)


def test_trig():
    """Test `numpy` trigonometric ufuncs."""
    nA = Symbol('A')
    v = numpy.pi / 3
    this = x(v, name=nA)
    that = numpy.sin(this)
    assert isinstance(that, oprattr.Operand)
    assert that == x(numpy.sin(v), name=Symbol('numpy.sin(A)'))
    that = numpy.cos(this)
    assert isinstance(that, oprattr.Operand)
    assert that == x(numpy.cos(v), name=Symbol('numpy.cos(A)'))
    that = numpy.tan(this)
    assert isinstance(that, oprattr.Operand)
    assert that == x(numpy.tan(v), name=Symbol('numpy.tan(A)'))


def test_sqrt():
    """Test `numpy.sqrt`."""
    this = x(4, name=Symbol('A'))
    that = numpy.sqrt(this)
    assert isinstance(that, oprattr.Operand)
    assert that == x(2, name=Symbol('numpy.sqrt(A)'))


def test_logs():
    """Test `numpy` logarithmic ufuncs."""
    nA = Symbol('A')
    v = 10.0
    this = x(v, name=nA)
    that = numpy.log(this)
    assert isinstance(that, oprattr.Operand)
    assert that == x(numpy.log(v), name=Symbol('numpy.log(A)'))
    that = numpy.log2(this)
    assert isinstance(that, oprattr.Operand)
    assert that == x(numpy.log2(v), name=Symbol('numpy.log2(A)'))
    that = numpy.log10(this)
    assert isinstance(that, oprattr.Operand)
    assert that == x(numpy.log10(v), name=Symbol('numpy.log10(A)'))


def test_squeeze():
    """Test `numpy.squeeze`."""
    nA = Symbol('A')
    nR = Symbol('numpy.squeeze(A)')
    v = numpy.array([[[1], [2], [3]]])
    this = x(v, name=nA)
    that = numpy.squeeze(this)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.squeeze(v), name=nR))
    that = numpy.squeeze(this, axis=0)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.squeeze(v, axis=0), name=nR))
    that = numpy.squeeze(this, axis=-1)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.squeeze(v, axis=-1), name=nR))


def test_mean():
    """Test `numpy.mean`."""
    nA = Symbol('A')
    nR = Symbol('numpy.mean(A)')
    v = numpy.arange(3*4*5).reshape(3, 4, 5)
    this = x(v, name=nA)
    that = numpy.mean(this)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.mean(v), name=nR))
    that = numpy.mean(this, axis=0)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.mean(v, axis=0), name=nR))
    that = numpy.mean(this, axis=-1)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.mean(v, axis=-1), name=nR))


def test_sum():
    """Test `numpy.sum`."""
    nA = Symbol('A')
    nR = Symbol('numpy.sum(A)')
    v = numpy.arange(3*4*5).reshape(3, 4, 5)
    this = x(v, name=nA)
    that = numpy.sum(this)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.sum(v), name=nR))
    that = numpy.sum(this, axis=0)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.sum(v, axis=0), name=nR))
    that = numpy.sum(this, axis=-1)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.sum(v, axis=-1), name=nR))


def test_cumsum():
    """Test `numpy.cumsum`."""
    nA = Symbol('A')
    nR = Symbol('numpy.cumsum(A)')
    v = numpy.arange(3*4*5).reshape(3, 4, 5)
    this = x(v, name=nA)
    that = numpy.cumsum(this)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.cumsum(v), name=nR))
    that = numpy.cumsum(this, axis=0)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.cumsum(v, axis=0), name=nR))
    that = numpy.cumsum(this, axis=-1)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.cumsum(v, axis=-1), name=nR))


def test_transpose():
    """Test `numpy.transpose`."""
    nA = Symbol('A')
    nR = Symbol('numpy.transpose(A)')
    v = numpy.arange(3*4*5).reshape(3, 4, 5)
    this = x(v, name=nA)
    that = numpy.transpose(this)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.transpose(v), name=nR))
    for axes in itertools.permutations(range(v.ndim)):
        that = numpy.transpose(this, axes=axes)
        assert isinstance(that, oprattr.Operand)
        expected = x(numpy.transpose(v, axes=axes), name=nR)
        assert numpy.array_equal(that, expected)


def test_gradient():
    """Test `numpy.gradient`."""
    nA = Symbol('A')
    nR = Symbol('numpy.gradient(A)')
    v = numpy.arange(3*4*5).reshape(3, 4, 5)
    this = x(v, name=nA)
    that = numpy.gradient(this)
    grad = numpy.gradient(v)
    for t, g in zip(that, grad):
        assert isinstance(t, oprattr.Operand)
        assert numpy.array_equal(t, x(g, name=nR))
    for axis in range(v.ndim):
        that = numpy.gradient(this, axis=axis)
        assert isinstance(that, oprattr.Operand)
        grad = numpy.gradient(v, axis=axis)
        assert numpy.array_equal(that, x(grad, name=nR))


def test_trapezoid():
    """Test `numpy.trapezoid`."""
    nA = Symbol('A')
    nR = Symbol('numpy.trapezoid(A)')
    v = numpy.arange(3*4*5).reshape(3, 4, 5)
    this = x(v, name=nA)
    that = numpy.trapezoid(this)
    assert isinstance(that, oprattr.Operand)
    assert numpy.array_equal(that, x(numpy.trapezoid(v), name=nR))
    for axis in range(v.ndim):
        that = numpy.trapezoid(this, axis=axis)
        assert isinstance(that, oprattr.Operand)
        trap = numpy.trapezoid(v, axis=axis)
        assert numpy.array_equal(that, x(trap, name=nR))




import collections.abc
import functools
import numbers

from numerical import operators
import numpy

from . import abstract
from . import mixins
from . import typeface
from ._operations import (
    unary,
    equality,
    ordering,
    additive,
    multiplicative,
)


T = typeface.TypeVar('T')


class Operand(abstract.Object[T], mixins.Numpy):
    """A concrete implementation of a real-valued object."""

    def __abs__(self):
        """Called for abs(self)."""
        return unary(operators.abs, self)

    def __pos__(self):
        """Called for +self."""
        return unary(operators.pos, self)

    def __neg__(self):
        """Called for -self."""
        return unary(operators.neg, self)

    def __eq__(self, other):
        """Called for self == other."""
        return equality(operators.eq, self, other)

    def __ne__(self, other):
        """Called for self != other."""
        return equality(operators.ne, self, other)

    def __lt__(self, other):
        """Called for self < other."""
        return ordering(operators.lt, self, other)

    def __le__(self, other):
        """Called for self <= other."""
        return ordering(operators.le, self, other)

    def __gt__(self, other):
        """Called for self > other."""
        return ordering(operators.gt, self, other)

    def __ge__(self, other):
        """Called for self >= other."""
        return ordering(operators.ge, self, other)

    def __add__(self, other):
        """Called for self + other."""
        return additive(operators.add, self, other)

    def __radd__(self, other):
        """Called for other + self."""
        return additive(operators.add, other, self)

    def __sub__(self, other):
        """Called for self - other."""
        return additive(operators.sub, self, other)

    def __rsub__(self, other):
        """Called for other - self."""
        return additive(operators.sub, other, self)

    def __mul__(self, other):
        """Called for self * other."""
        return multiplicative(operators.mul, self, other)

    def __rmul__(self, other):
        """Called for other * self."""
        return multiplicative(operators.mul, other, self)

    def __truediv__(self, other):
        """Called for self / other."""
        return multiplicative(operators.truediv, self, other)

    def __rtruediv__(self, other):
        """Called for other / self."""
        return multiplicative(operators.truediv, other, self)

    def __floordiv__(self, other):
        """Called for self // other."""
        return multiplicative(operators.floordiv, self, other)

    def __rfloordiv__(self, other):
        """Called for other // self."""
        return multiplicative(operators.floordiv, other, self)

    def __mod__(self, other):
        """Called for self % other."""
        return multiplicative(operators.mod, self, other)

    def __rmod__(self, other):
        """Called for other % self."""
        return multiplicative(operators.mod, other, self)

    def __pow__(self, other):
        """Called for self ** other."""
        if isinstance(other, numbers.Real):
            return multiplicative(operators.pow, self, other)
        return NotImplemented

    def __rpow__(self, other):
        """Called for other ** self."""
        return NotImplemented

    def __array__(self, *args, **kwargs):
        """Called for numpy.array(self)."""
        return numpy.array(self._data, *args, **kwargs)

    def _apply_ufunc(self, ufunc, method, *args, **kwargs):
        if ufunc in (numpy.equal, numpy.not_equal):
            # NOTE: We are probably here because the left operand is a
            # `numpy.ndarray`, which would otherwise take control and return the
            # pure `numpy` result.
            f = getattr(ufunc, method)
            return equality(f, *args)
        return super()._apply_ufunc(ufunc, method, *args, **kwargs)


@Operand.implementation(numpy.array_equal)
def array_equal(
    x: numpy.typing.ArrayLike,
    y: numpy.typing.ArrayLike,
    **kwargs
) -> bool:
    """Called for numpy.array_equal(x, y)"""
    return numpy.array_equal(numpy.array(x), numpy.array(y), **kwargs)


@Operand.implementation(numpy.gradient)
def gradient(x: Operand[T], *args, **kwargs):
    """Called for numpy.gradient(x)."""
    data = numpy.gradient(x._data, *args, **kwargs)
    meta = {}
    for key, value in x._meta.items():
        try:
            v = numpy.gradient(value, **kwargs)
        except TypeError as exc:
            raise TypeError(
                "Cannot compute numpy.gradient(x)"
                f" because metadata attribute {key!r}"
                " does not support this operation"
            ) from exc
        else:
            meta[key] = v
    if isinstance(data, (list, tuple)):
        r = [type(x)(array, **meta) for array in data]
        if isinstance(data, tuple):
            return tuple(r)
        return r
    return type(x)(data, **meta)


def wrapnumpy(f: collections.abc.Callable):
    """Implement a numpy function for objects with metadata."""
    @functools.wraps(f)
    def method(x: Operand[T], **kwargs):
        """Apply a numpy function to x."""
        data = f(x._data, **kwargs)
        meta = {}
        for key, value in x._meta.items():
            try:
                v = f(value, **kwargs)
            except TypeError as exc:
                raise TypeError(
                    f"Cannot compute numpy.{f.__qualname__}(x)"
                    f" because metadata attribute {key!r}"
                    " does not support this operation"
                ) from exc
            else:
                meta[key] = v
        return type(x)(data, **meta)
    return method


_OPERAND_UFUNCS = (
    numpy.sqrt,
    numpy.sin,
    numpy.cos,
    numpy.tan,
    numpy.log,
    numpy.log2,
    numpy.log10,
)


_OPERAND_FUNCTIONS = (
    numpy.squeeze,
    numpy.mean,
    numpy.sum,
    numpy.cumsum,
    numpy.transpose,
    numpy.trapezoid,
)


for f in _OPERAND_UFUNCS + _OPERAND_FUNCTIONS:
    Operand.implement(f, wrapnumpy(f))





"""
This script generates and echos exceptions related to operations on instances of
the `Operand` class. It is meant as a supplement to rigorous tests.
"""

import oprattr

x = oprattr.Operand(1, name='A')
y = oprattr.Operand(2, name='B')

cases = (
    (oprattr.operators.lt, x, y),
    (oprattr.operators.lt, x, 2),
    (oprattr.operators.lt, 2, x),
    (oprattr.operators.add, x, y),
    (oprattr.operators.add, x, 2),
    (oprattr.operators.add, 2, y),
    (oprattr.operators.abs, x),
    (oprattr.operators.mul, x, 'y'),
    (oprattr.operators.mul, 'x', y),
    (oprattr.operators.pow, x, 2)
)

for f, *args in cases:
    try:
        f(*args)
    except Exception as exc:
        print(f"Caught {type(exc).__qualname__}: {exc}")
    else:
        strargs = ', '.join(str(arg) for arg in args)
        print(f"Calling {f} on {strargs} did not raise an exception")




from numerical import operators

from .abstract import Quantity


class MetadataError(TypeError):
    """A metadata-related TypeError occurred."""

    def __init__(
        self,
        f: operators.Operator,
        *args,
        error: str | None = None,
        key: str | None = None,
    ) -> None:
        super().__init__(*args)
        self._f = f
        self._error = error
        self._key = key

    def __str__(self):
        """Called when handling the exception."""
        types = [type(arg) for arg in self.args]
        return _build_error_message(
            self._f,
            *types,
            error=self._error,
            key=self._key,
        )


def _build_error_message(
    f: operators.Operator,
    *types: type,
    error: str | None = None,
    key: str | None = None,
) -> str:
    """Helper for `_raise_metadata_exception`.
    
    This function should avoid raising an exception if at all possible, and
    instead return the default error message, since it is already being called
    as the result of an error elsewhere.
    """
    errmsg = f"Cannot compute {f}"
    errstr = error.lower() if isinstance(error, str) else ''
    if errstr == 'unequal':
        return f"{errmsg} between objects with unequal metadata"
    if errstr in {'non-empty', 'nonempty'}:
        if len(types) == 2:
            a, b = types
            endstr = "because {} has metadata"
            if issubclass(a, Quantity):
                return f"{errmsg} between {a} and {b} {endstr.format(str(a))}"
            if issubclass(b, Quantity):
                return f"{errmsg} between {a} and {b} {endstr.format(str(b))}"
    if errstr == 'type':
        if key is None:
            keystr = "a metadata attribute"
        else:
            keystr = f"metadata attribute {key!r}"
        midstr = f"because {keystr}"
        endstr = "does not support this operation"
        if len(types) == 1:
            return f"{errmsg} of {types[0]} {midstr} {endstr}"
        if len(types) == 2:
            a, b = types
            return f"{errmsg} between {a} and {b} {midstr} {endstr}"
    return errmsg


def unary(f: operators.Operator, a):
    """Compute the unary operation f(a)."""
    if isinstance(a, Quantity):
        meta = {}
        for key, value in a._meta.items():
            try:
                v = f(value)
            except TypeError as exc:
                raise MetadataError(f, a, error='type', key=key) from exc
            else:
                meta[key] = v
        return type(a)(f(a._data), **meta)
    return f(a)


def equality(f: operators.Operator, a, b):
    """Compute the equality operation f(a, b)."""
    x = a._data if isinstance(a, Quantity) else a
    y = b._data if isinstance(b, Quantity) else b
    fxy = f(x, y)
    try:
        iter(fxy)
    except TypeError:
        r = bool(fxy)
    else:
        r = all(fxy)
    isne = f(1, 2)
    if isinstance(a, Quantity) and isinstance(b, Quantity):
        if a._meta != b._meta:
            return isne
        return r
    if isinstance(a, Quantity):
        if not a._meta:
            return r
        return isne
    if isinstance(b, Quantity):
        if not b._meta:
            return r
        return isne
    return r


def ordering(f: operators.Operator, a, b):
    """Compute the ordering operation f(a, b)."""
    if isinstance(a, Quantity) and isinstance(b, Quantity):
        if a._meta == b._meta:
            return f(a._data, b._data)
        raise MetadataError(f, a, b, error='unequal') from None
    if isinstance(a, Quantity):
        if not a._meta:
            return f(a._data, b)
        raise MetadataError(f, a, b, error='non-empty') from None
    if isinstance(b, Quantity):
        if not b._meta:
            return f(a, b._data)
        raise MetadataError(f, a, b, error='non-empty') from None
    return f(a, b)


def additive(f: operators.Operator, a, b):
    """Compute the additive operation f(a, b)."""
    if isinstance(a, Quantity) and isinstance(b, Quantity):
        if a._meta == b._meta:
            return type(a)(f(a._data, b._data), **a._meta)
        raise MetadataError(f, a, b, error='unequal') from None
    if isinstance(a, Quantity):
        if not a._meta:
            return type(a)(f(a._data, b))
        raise MetadataError(f, a, b, error='non-empty') from None
    if isinstance(b, Quantity):
        if not b._meta:
            return type(b)(f(a, b._data))
        raise MetadataError(f, a, b, error='non-empty') from None
    return f(a, b)


def multiplicative(f: operators.Operator, a, b):
    """Compute the multiplicative operation f(a, b)."""
    if isinstance(a, Quantity) and isinstance(b, Quantity):
        keys = set(a._meta) & set(b._meta)
        meta = {}
        for key in keys:
            try:
                v = f(a._meta[key], b._meta[key])
            except TypeError as exc:
                raise MetadataError(f, a, b, error='type', key=key) from exc
            else:
                meta[key] = v
        for key, value in a._meta.items():
            if key not in keys:
                meta[key] = value
        for key, value in b._meta.items():
            if key not in keys:
                meta[key] = value
        return type(a)(f(a._data, b._data), **meta)
    if isinstance(a, Quantity):
        meta = {}
        for key, value in a._meta.items():
            try:
                v = f(value, b)
            except TypeError as exc:
                raise MetadataError(f, a, b, error='type', key=key) from exc
            else:
                meta[key] = v
        return type(a)(f(a._data, b), **meta)
    if isinstance(b, Quantity):
        meta = {}
        for key, value in b._meta.items():
            try:
                v = f(a, value)
            except TypeError as exc:
                raise MetadataError(f, a, b, error='type', key=key) from exc
            else:
                meta[key] = v
        return type(b)(f(a, b._data), **meta)
    return f(a, b)





"""
Support for type annotations.

This module provides a single interface to type annotations, including those
that are not defined by the operative Python version and those that this package
prefers to use from future versions.

Examples
--------
* Suppose `BestType` is available in the `typing` module starting with Python
  version 3.X and is available in the `typing_extensions` module for earlier
  versions. If the user is running with Python version <3.X, this module will
  import `BestType` from `typing_extensions`. Otherwise, it will import
  `BestType` from `typing`.
* Support `UpdatedType` is available in the `typing` module for the user's
  version of Python, but this package wishes to take advantage of updates since
  that version. This module will automatically import the version from
  `typing_extensions`.
"""

import typing
import typing_extensions


__all__ = ()

EXTENDED = [
    'Protocol',
]

def __getattr__(name: str) -> type:
    """Get a built-in type annotation."""
    if name in EXTENDED:
        return getattr(typing_extensions, name)
    try:
        attr = getattr(typing, name)
    except AttributeError:
        attr = getattr(typing_extensions, name)
    return attr






import collections.abc
import numbers

import numerical
import numpy.typing

from . import typeface


DataType = typeface.TypeVar(
    'DataType',
    int,
    float,
    numbers.Number,
    numpy.number,
    numpy.typing.ArrayLike,
    numpy.typing.NDArray,
)


@typeface.runtime_checkable
class Quantity(numerical.Quantity[DataType], typeface.Protocol):
    """Protocol for numerical objects with metadata."""

    _meta: collections.abc.Mapping[str, typeface.Any]


class Object(numerical.Real, typeface.Generic[DataType]):
    """A real-valued object with metadata attributes."""

    def __init__(
        self,
        __data: DataType,
        **metadata,
    ) -> None:
        if not isinstance(__data, numerical.Real):
            raise TypeError("Data input to Object must be real-valued")
        self._data = __data
        self._meta = metadata

    def __repr__(self):
        """Called for repr(self)."""
        try:
            datastr = numpy.array2string(
                self._data,
                separator=", ",
                threshold=6,
                edgeitems=2,
                prefix=f"{self.__class__.__qualname__}(",
                suffix=")"
            )
        except Exception:
            datastr = str(self._data)
        metastr = "metadata={" + ", ".join(f"{k!r}" for k in self._meta) + "}"
        return f"{self.__class__.__qualname__}({datastr}, {metastr})"




import collections.abc
import numbers

import numpy

from . import abstract
from . import typeface


T = typeface.TypeVar('T')


class Real:
    """Mixin for adding basic real-valued operator support."""

    def __abs__(self):
        return self

    def __pos__(self):
        return self

    def __neg__(self):
        return self

    def __eq__(self, other):
        return False

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return (self < other) and (self == other)

    def __gt__(self, other):
        return not (self <= other)

    def __ge__(self, other):
        return not (self < other)

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __rmul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __rtruediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __mod__(self, other):
        return self

    def __rmod__(self, other):
        return self

    def __pow__(self, other):
        return self

    def __rpow__(self, other):
        return self


UserFunction = collections.abc.Callable[..., T]


class Numpy:
    """Mixin for adding support for `numpy` functions to numeric objects.
    
    Classes that inherit from this class may implement support for `numpy`
    universal functions ("ufuncs"; e.g., `numpy.sqrt`) by overloading
    `_apply_ufunc`, and may implement support for `numpy` public functions
    (e.g., `numpy.squeeze`) by overloading `_apply_function` and registering
    individual function implementations via `implementation`.

    It is important to note that the use cases of this class extend beyond
    array-like objects. Both single- and multi-valued objects can benefit from
    implementing support for `numpy` universal and public functions. For
    example, it is possible to apply `numpy.sqrt` to both a real number and an
    array

    >>> numpy.sqrt(4)
    2.0
    >>> numpy.sqrt([4, 9])
    array([2., 3.])

    Even the trivial application of `numpy.mean` to a real number is defined:

    >>> numpy.mean(2.5)
    2.5

    Notes
    -----
    - This class does not inherit from `numpy.lib.mixins.NDArrayOperatorsMixin`,
      which implements most of the built-in Python numeric operators via
      `__array_ufunc__`, because it assumes that subclasses independently
      implement those methods.
    """

    def __init_subclass__(cls):
        cls._UFUNC_TYPES |= {cls}
        cls._FUNCTION_TYPES |= {cls}
        cls._FUNCTIONS = {}

    _UFUNC_TYPES = {
        numpy.ndarray,
        numbers.Number,
        list,
        abstract.Quantity,
    }

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        """Provide support for `numpy` universal functions.

        See https://numpy.org/doc/stable/reference/arrays.classes.html for more
        information on use of this special method.

        Notes
        -----
        - This method first ensures that the input types (as well as the type of
          `out`, if passed via keyword) are supported types. It then checks for
          a custom implementation of `ufunc`. If there is a custom
          implementation, this method applies it and returns the result. If
          there is no custom implementation, this method passes control to
          `_apply_ufunc`, to allow subclass customization.
        - See `implementation` for additional guidance on custom
          implementations.

        See Also
        --------
        `implementation`
            Class method for registering custom ufunc implementations.

        `_apply_ufunc`
            Instance method that allows custom handling of ufuncs corresponding
            to standard Python numerical operators.
        """
        out = kwargs.get('out', ())
        accepted = tuple(self._UFUNC_TYPES)
        if not all(isinstance(x, accepted) for x in args + out):
            return NotImplemented
        if out:
            kwargs['out'] = tuple(
                x._data if isinstance(x, abstract.Quantity)
                else x for x in out
            )
        if self._implements(ufunc):
            operator = self._FUNCTIONS[ufunc]
            return operator(*args, **kwargs)
        return self._apply_ufunc(ufunc, method, *args, **kwargs)

    def _apply_ufunc(self, ufunc, method, *args, **kwargs):
        """Apply a `numpy` universal function (a.k.a "ufunc") to data.

        Notes
        -----
        - Subclasses that wish to customize support for ufuncs should overload
          this method instead of `__array_ufunc__`.
        - Subclasses should prefer to define custom implementations of specific
          universal functions and register each via `implementation`, rather
          than implementing function-specific logic in this method, since
          `__array_ufunc__` will check for a custom implementation of a given
          function before calling this method.
        - The default implementation of this method applies the given ufunc to
          real-valued data and directly returns the `numpy` result, without
          attempting to create a new instance of the custom subclass.

        See Also
        --------
        `implementation`
            Class method for registering custom ufunc implementations.

        `__array_ufunc__`
            The entry point for `numpy` universal functions.
        """
        operator = getattr(ufunc, method)
        values = self._get_numpy_args(args)
        try:
            data = operator(*values, **kwargs)
        except TypeError as err:
            raise TypeError(
                f"Unable to apply {ufunc} to {args}"
            ) from err
        if method != 'at':
            return data

    _FUNCTION_TYPES = {
        numpy.ndarray,
        abstract.Object,
     } | set(numpy.ScalarType)

    def __array_function__(self, func, types, args, kwargs):
        """Provide support for functions in the `numpy` public API.

        See https://numpy.org/doc/stable/reference/arrays.classes.html for more
        information of use of this special method. The implementation shown here
        is a combination of the example on that page and code from the
        definition of `EncapsulateNDArray.__array_function__` in
        https://github.com/dask/dask/blob/main/dask/array/tests/test_dispatch.py

        Notes
        -----
        - This method first checks that all `types` are in
          `self._FUNCTION_TYPES`, thereby allowing subclasses that don't
          override `__array_function__` to handle objects of this type. It then
          checks for a custom implementation of `func`. If there is a custom
          implementation, this method applies it and returns the result. If
          there is no custom implementation, this method passes control to
          `_apply_function`, to allow subclass customization.
        - See `implementation` for additional guidance on custom
          implementations.

        See Also
        --------
        `implementation`
            Class method for registering custom function implementations.

        `_apply_function`
            Instance method that allows custom handling of `numpy` public
            functions when there is no registered custom implementation.
        """
        accepted = tuple(self._FUNCTION_TYPES)
        if not all(issubclass(ti, accepted) for ti in types):
            return NotImplemented
        if self._implements(func):
            return self._FUNCTIONS[func](*args, **kwargs)
        return self._apply_function(func, types, args, kwargs)

    def _apply_function(self, func, types, args, kwargs):
        """Apply a function in the `numpy` public API.

        Notes
        -----
        - Subclasses that wish to customize support for public functions should
          overload this method instead of `__array_function__`.
        - Subclasses should prefer to define custom implementations of specific
          public functions and register each via `implementation`, rather than
          implementing function-specific logic in this method, since
          `__array_function__` will check for a custom implementation of a given
          function before calling this method.
        - The default implementation calls `_get_numpy_array` for access to
          real-valued data via an instance of `numpy.ndarray`, `_get_numpy_args`
          to convert `args` to appropriate operands, and `_get_numpy_types` to
          extract appropriate operand types. Subclasses may choose to overload
          any of those individual methods instead of overloading this method.

        See Also
        --------
        `implementation`
            Class method for registering custom ufunc implementations.

        `__array_function__`
            The entry point for `numpy` public functions.
        """
        array = self._get_numpy_array()
        if array is None:
            return NotImplemented
        if not isinstance(array, numpy.ndarray):
            raise TypeError(
                f"{self.__class__.__qualname__}._get_numpy_array"
                " did not return a numpy.ndarray"
            ) from None
        args = self._get_numpy_args(args)
        types = self._get_numpy_types(types)
        return array.__array_function__(func, types, args, kwargs)

    def _get_numpy_array(self) -> numpy.typing.NDArray | None:
        """Convert the data interface to an array for `numpy` mixin methods.
        
        Notes
        -----
        - This method allows subclass implementations to control how they
          convert their data interface to a `numpy.ndarray` for use with `numpy`
          public functions.
        - Returning `None` from this method will cause `_apply_function` to
          return `NotImplemented`.
        - The default implementation unconditionally returns `None`.
        """
        return

    def _get_numpy_args(self, args):
        """Convert `args` to operands of a `numpy` function.

        This method will call `~_get_arg_data` on each member of `args` in order
        to build a `tuple` of suitable operands. Subclasses may overload
        `~_get_arg_data` to customize access to their data attribute.
        """
        return tuple(self._get_arg_data(arg) for arg in args)

    def _get_arg_data(self, arg):
        """Convert `arg` to an operand of a `numpy` function.

        See Also
        --------
        `~_get_numpy_args`
            The method that calls this method in a loop.

        Notes
        -----
        - This method allows a subclass to customize how `numpy` functions
          access its data attribute.
        - The default implementation will return the `data` attribute of a of
          `arg` if `arg` is an instance of the base object class; otherwise, it
          will return the unmodified argument.
        """
        if isinstance(arg, abstract.Quantity):
            return arg._data
        return arg

    def _get_numpy_types(self, types):
        """Extract appropriate types for a `numpy` function.
        
        Notes
        -----
        - This method allows subclasses to restrict the object types that they
          pass to `numpy` public functions via `_apply_function`.
        - The default implementation returns a tuple that contains all types
          except for subtypes of `~_types.Quantity`.
        """
        return tuple(
            ti for ti in types
            if not issubclass(ti, abstract.Object)
        )

    @classmethod
    def _implements(cls, operation: collections.abc.Callable):
        """True if this class defines a custom implementation for `operation`.
        
        This is a helper methods that gracefully handles the case in which a
        subclass does not support custom operator implementations.
        """
        try:
            result = operation in cls._FUNCTIONS
        except TypeError:
            return False
        return result

    _FUNCTIONS: dict[str, collections.abc.Callable]=None
    """Internal collection of custom `numpy` function implementations."""

    @classmethod
    def implementation(cls, numpy_function: collections.abc.Callable, /):
        """Register a custom implementation of this `numpy` function.

        Parameters
        ----------
        numpy_function : callable
            The `numpy` universal or public function to implement.

        Notes
        -----
        - Users may register `numpy` universal functions (a.k.a. ufuncs;
          https://numpy.org/doc/stable/reference/ufuncs.html) as well as
          functions in the public `numpy` API (e.g., `numpy.mean`). This may be
          important if, for example, a class needs to implement a custom version
          of `numpy.sqrt`, which is a ufunc.
        - See https://numpy.org/doc/stable/reference/arrays.classes.html for the
          suggestion on which this method is based.

        Examples
        --------
        Overload `numpy.mean` for an existing class called `Array` with a
        version that accepts no keyword arguments:

        ```
            @Array.implementation(numpy.mean)
            def mean(a: Array, **kwargs) -> Array:
                if kwargs:
                    msg = "Cannot pass keywords to numpy.mean with Array" raise
                    TypeError(msg)
                return numpy.sum(a) / len(a)
        ```

        This will compute the mean of the underlying data when called with no
        arguments, but will raise an exception when called with arguments:

            >>> v = Array([[1, 2], [3, 4]])
            >>> numpy.mean(v)
            5.0
            >>> numpy.mean(v, axis=0)
            ...
            TypeError: Cannot pass keywords to numpy.mean with Array

        See Also
        --------
        `~implements`
        """
        if not callable(numpy_function):
            raise TypeError(
                "The target operation of a custom numpy implementation"
                " must be callable"
            ) from None
        def decorator(user_function: UserFunction):
            if cls._FUNCTIONS is None:
                raise NotImplementedError(
                    f"Type {cls} does not support custom implementations"
                    " of numpy functions"
                ) from None
            cls._FUNCTIONS[numpy_function] = user_function
            return user_function
        return decorator

    @classmethod
    def implement(
        cls,
        numpy_function: collections.abc.Callable,
        user_function: UserFunction,
        /,
    ) -> None:
        """Implement a `numpy` function via a given user function.

        This method serves as an alternative to the class method
        `implementation`, which is primarily meant to be used as a decorator.
        This method allows the user to directly associate a custom
        implementation with the target `numpy` function.

        Parameters
        ----------
        numpy_function : callable
            The `numpy` universal or public function to implement.

        user_function: callable
            The custom implementation to associate with `numpy_function`.

        Examples
        --------
        Here is an alternative to the `~implementation` example usage:

        ```
            def mean(a: Array, **kwargs) -> Array:
                if kwargs:
                    msg = "Cannot pass keywords to numpy.mean with Array" raise
                    TypeError(msg)
                return numpy.sum(a) / len(a)

            Array.implement(numpy.mean, mean)
        ```

        However, a more useful application may be to associate multiple `numpy`
        functions with a single custom implementation:

        ```
            def trig(f: numpy.ufunc):
                def method(a: Array):
                    ... # custom implementation
                return method

            for f in {numpy.sin, numpy.cos, numpy.tan}:
                Array.implement(f, trig(f))
        ```

        See Also
        --------
        `~implementation`
        """
        if not callable(numpy_function):
            raise TypeError(
                "The target operation of a custom numpy implementation"
                " must be callable"
            ) from None
        if cls._FUNCTIONS is None:
            raise NotImplementedError(
                f"Type {cls} does not support custom implementations"
                " of numpy functions"
            ) from None
        cls._FUNCTIONS[numpy_function] = user_function

